

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <cutil.h>
#include <windows.h>
#include <assert.h>
#include <template_kernel.cu>

void host_side( int argc, char** argv);
void device_side_1( int argc, char** argv);
void device_side_2( int argc, char** argv);

#define NUMTHREADS		256*4
#define	PRINT_RESULT

int
main( int argc, char** argv) 
{
	printf("Coalesced Memory Test\n");

	CUDA_SAFE_CALL( cudaSetDevice(0) );

	//device_side_1( argc, argv);		// for measuring uncoalesced memory access pattern
	device_side_2( argc, argv);			// for measuring coalesced memory access pattern

    CUT_EXIT(argc, argv);
}

void 
host_side( int argc, char** argv) 
{
	
}

void
device_side_1( int argc, char** argv) 
{
	LARGE_INTEGER liCounter1, liCounter2, liFrequency;
	int cnt_i, cnt_j;
	unsigned int *hOpA, *hOpB, *hRet;
	unsigned int *dOpA, *dOpB, *dRet;
	dim3 grid(1, 1);
	dim3 blocks(1, 16);

	hOpA = (unsigned int*)calloc(NUMTHREADS, sizeof(unsigned int));
	hOpB = (unsigned int*)calloc(NUMTHREADS, sizeof(unsigned int));
	hRet = (unsigned int*)calloc(NUMTHREADS, sizeof(unsigned int));
	assert(hOpA);
	assert(hOpB);
	assert(hRet);
	CUDA_SAFE_CALL(cudaMalloc((void**)&dOpA, NUMTHREADS*sizeof(unsigned int)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&dOpB, NUMTHREADS*sizeof(unsigned int)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&dRet, NUMTHREADS*sizeof(unsigned int)));

	for(cnt_i = 0; cnt_i < NUMTHREADS; cnt_i++)
	{
		hOpB[cnt_i] = hOpA[cnt_i] = cnt_i;
	}

	CUDA_SAFE_CALL(cudaMemcpy(dOpA, hOpA, NUMTHREADS*sizeof(unsigned int), cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy(dOpB, hOpB, NUMTHREADS*sizeof(unsigned int), cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy(dRet, hRet, NUMTHREADS*sizeof(unsigned int), cudaMemcpyHostToDevice));

	QueryPerformanceFrequency(&liFrequency);		// 1ʴ Ǵ īƮ  Ѵ

	// no bank conflict
	QueryPerformanceCounter(&liCounter1); 
	for(cnt_i = 0; cnt_i < 10000; cnt_i++)
	{
		UnCoalscedMemTestKernel<<<grid, blocks>>>(dRet, dOpA, dOpB);
	}
	QueryPerformanceCounter(&liCounter2);

	CUDA_SAFE_CALL(cudaMemcpy(hRet, dRet, NUMTHREADS*sizeof(unsigned int), cudaMemcpyDeviceToHost));

	printf("\n[GPU] Uncoalesed Access: %lf \n", ((double)(liCounter2.QuadPart-liCounter1.QuadPart)/(double)liFrequency.QuadPart));

#if defined PRINT_RESULT
	for(cnt_i = 0; cnt_i < NUMTHREADS; cnt_i++)
	{
		if(cnt_i && !(cnt_i % 8))
			printf("\n");
		printf("%0.6d ", hRet[cnt_i]);
	}
#endif

	if(hOpA)
		free(hOpA);
	if(hOpB)
		free(hOpB);
	if(hRet)
		free(hRet);
}

void
device_side_2( int argc, char** argv) 
{
	LARGE_INTEGER liCounter1, liCounter2, liFrequency;
	unsigned int *hOpA, *hOpB, *hRet;
	unsigned int *dOpA, *dOpB, *dRet;
	dim3 grid(1, 1);
	dim3 blocks(1, 16);
	int cnt_i;

	hOpA = (unsigned int*)calloc(NUMTHREADS, sizeof(unsigned int));
	hOpB = (unsigned int*)calloc(NUMTHREADS, sizeof(unsigned int));
	hRet = (unsigned int*)calloc(NUMTHREADS, sizeof(unsigned int));
	assert(hOpA);
	assert(hOpB);
	assert(hRet);
	CUDA_SAFE_CALL(cudaMalloc((void**)&dOpA, NUMTHREADS*sizeof(unsigned int)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&dOpB, NUMTHREADS*sizeof(unsigned int)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&dRet, NUMTHREADS*sizeof(unsigned int)));

	for(cnt_i = 0; cnt_i < NUMTHREADS; cnt_i++)
	{
		hOpB[cnt_i] = hOpA[cnt_i] = cnt_i;
	}

	CUDA_SAFE_CALL(cudaMemcpy(dOpA, hOpA, NUMTHREADS*sizeof(unsigned int), cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy(dOpB, hOpB, NUMTHREADS*sizeof(unsigned int), cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy(dRet, hRet, NUMTHREADS*sizeof(unsigned int), cudaMemcpyHostToDevice));

	QueryPerformanceFrequency(&liFrequency);		// 1ʴ Ǵ īƮ  Ѵ

	// bank conflict
	QueryPerformanceCounter(&liCounter1); 
	for(cnt_i = 0; cnt_i < 10000; cnt_i++)
	{
		CoalscedMemTestKernel<<<grid, blocks>>>(dRet, dOpA, dOpB);
	}
	QueryPerformanceCounter(&liCounter2);

	CUDA_SAFE_CALL(cudaMemcpy(hRet, dRet, NUMTHREADS*sizeof(unsigned int), cudaMemcpyDeviceToHost));

	printf("\n[GPU] Coalesed Access: %lf \n", ((double)(liCounter2.QuadPart-liCounter1.QuadPart)/(double)liFrequency.QuadPart));

	#if defined PRINT_RESULT
	for(cnt_i = 0; cnt_i < NUMTHREADS; cnt_i++)
	{
		if(cnt_i && !(cnt_i % 8))
			printf("\n");
		printf("%0.6d ", hRet[cnt_i]);
	}
#endif

	if(hOpA)
		free(hOpA);
	if(hOpB)
		free(hOpB);
	if(hRet)
		free(hRet);
}

// EOF
